/*
 *  BINtegrity aka Syswall
 *
 */

#define CAT(X,Y) X##_##Y
#define CHECK_SYSCALL(X) CAT(check_syscall,X)

int CHECK_SYSCALL(CL) (unsigned arg1, unsigned arg2) {
    DPRINT("Syscall by %s pid %d, tgid %d\n", current->comm, current->pid, current->tgid);

    struct pt_regs *ptregs = task_pt_regs(current);
    unsigned argv[6] = {0};
    unsigned argc = 4;
    int lr_kill = 0;
#ifdef CONFIG_ARM
    unsigned syscall_number = ptregs->ARM_r7;
    unsigned return_address = ptregs->ARM_pc;
    unsigned subreturn_address = ptregs->ARM_lr;
    argv[0] = ptregs->ARM_r0;
    argv[1] = ptregs->ARM_r1;
    argv[2] = ptregs->ARM_r2;
    argv[3] = ptregs->ARM_r3;
    argv[4] = ptregs->ARM_r4;
    argv[5] = ptregs->ARM_r5;
#endif
#ifdef CONFIG_MIPS
    unsigned syscall_number = ptregs->regs[2];
    unsigned return_address = ptregs->cp0_epc;
    unsigned subreturn_address = ptregs->regs[31];
    argv[0] = ptregs->regs[4];
    argv[1] = ptregs->regs[5];
    argv[2] = ptregs->regs[6];
    argv[3] = ptregs->regs[7];
    argv[4] = arg1;
    argv[5] = arg2;
    argc = (unsigned)(sys_call_table[(syscall_number - __NR_Linux) * 2 + 1]);
    argc = (argc + (5 << 2)) >> 2;
    if (argc > 4) argc = 4;
#endif
    
    DPRINT("Syscall #%u, argc %u argv %#x, %#x, %#x, %#x\n", syscall_number, argc, argv[0], argv[1], argv[2], argv[3]);
    DPRINT("Return address: %#x, subreturn address: %#x\n", return_address, subreturn_address);

    struct vm_area_struct *vm_area = current->mm->mmap;

    struct hashmap_t *hm = 0;
    if (vm_area->vm_start != 0x1000) {
        
        down_write(&current->mm->mmap_sem);
        int retval = do_mmap_pgoff(0, 0x1000, HASHMAP_MM, PROT_READ|PROT_WRITE, MAP_ANONYMOUS|MAP_PRIVATE, 0);
        
        up_write(&current->mm->mmap_sem);
        if (retval != 0x1000) {
            WARN("Hashmap memory alloc failed!\n");
            return -1;
        }

        hm = (struct hashmap_t*) retval;
        int hret = hashcreate(hm, 0, sizeof(struct vetmap_t), (void*)(retval + sizeof(struct hashmap_t)));
        if (!hret) {
            WARN("Hashmap create failed\n");
            return -1;
        }
        hm->dlopen_file = NULL;
        DPRINT("\tHashmap = %#x hm->size = %d element size = %d mm = %d\n", hm, hm->hm_size, hm->element_size, HASHMAP_MM);
    }
    else {
        hm = (struct hashmap_t*) vm_area->vm_start;
        vm_area = vm_area->vm_next;
    }

    if (!hm->hot) {
        vm_area = current->mm->mmap;
        do {
            if (vm_area->vm_start <= return_address && return_address < vm_area->vm_end && vm_area->vm_file && (vm_area->vm_flags & VM_EXECUTABLE || vm_area->vm_flags & VM_EXEC)) {
                if (strncmp(vm_area->vm_file->f_path.dentry->d_name.name, LINKER, LINKER_LEN)) {
                    DPRINT("\tLinker is done, creating trusted code base\n");
                    fill_vma(hm->vma, &hm->vma_num, current->mm->mmap);
                    hm->hot = 1;
                    break;
                } else {
                    DPRINT("\tSyscall issued by linker, skipping\n");
                    return 0;
                }
            }
            vm_area = vm_area->vm_next;
        } while (vm_area != NULL);
    }

    if (hm->dlopen_file) {
        vm_area = current->mm->mmap;
        do {
            if (vm_area->vm_file && vm_area->vm_file->f_path.dentry->d_inode->i_ino == hm->dlopen_file->f_path.dentry->d_inode->i_ino) {
                DPRINT("\tAdding dlopen mapped file\n");
                add_single_vma(hm->vma, &hm->vma_num, vm_area);
                break;
            }
            vm_area = vm_area->vm_next;
        } while (vm_area != NULL);
        hm->dlopen_file = NULL;
    }

    unsigned syscall_offset = 0;
    struct file *syscall_file = NULL;
    unsigned char *syscall_dynamic = 0;
    unsigned library_offset = 0;
    struct file *library_file = NULL;
    unsigned char *library_dynamic = 0;

    struct vm_area_fix_t *vma = (struct vm_area_fix_t*) hm->vma;
    for (int i = 0; i < hm->vma_num; i++) {
        if (!syscall_file && vma[i].start <= return_address && return_address < vma[i].end) {
            syscall_file = vma[i].f;
            syscall_offset = return_address - vma[i].start;
            syscall_dynamic = vma[i].dynamic;
#ifdef CONFIG_ARM
            if (vma[i].start == 0xb0001000) {syscall_offset += 0x1000;}
#endif
        }
        if (!library_file && vma[i].start <= subreturn_address && subreturn_address < vma[i].end) {
            library_file = vma[i].f;
            library_offset = subreturn_address - vma[i].start;
            library_dynamic = vma[i].dynamic;
#ifdef CONFIG_ARM              
            if (vma[i].start == 0xb0001000) {library_offset += 0x1000;}
#endif
        }
        if (syscall_file && library_file)
            break;
    }

    if (!syscall_file) {
        WARN_BASIC_INFO();
        force_sig(SIGKILL, current);
        return -1;
    }

    if (!library_file) {
        WARN_BASIC_INFO();
        lr_kill = 1;
    }

#if CL == ORIGIN
    return 0;
#endif

    unsigned syscall_stored = 0;
    unsigned int p = hashget(hm, return_address);
    struct vetmap_t *syscall_stored_vm = 0;
tryagain:
    syscall_stored_vm = hm->hm;
    syscall_stored_vm = &(syscall_stored_vm[p]);

    if (syscall_stored_vm->addr == return_address)
        syscall_stored = syscall_stored_vm->syscall_number;
#ifdef COLLISION_FREE
    else if (!syscall_stored_vm->addr) {
        syscall_stored_vm->ts = jiffies;
    }
    else {
        p = (p + 1) % HASHMAP_SIZE;
        goto tryagain;
    }
#endif

    if (syscall_number == __NR_mmap || syscall_number == __NR_mmap2) {
        if (!strncmp(syscall_file->f_path.dentry->d_name.name, LINKER, LINKER_LEN)) {
    #ifdef CONFIG_MIPS
            syscall_stored_vm->badlr = 1;
    #endif    
            if (!(argv[3] & MAP_ANONYMOUS) && argv[4]) {
                DPRINT("\tLinker dlopen is mapping more memory for fd %u\n", argv[4]);
                hm->dlopen_file = fget(argv[4]);
                if (hm->dlopen_file)
                    DPRINT("dlopen_file is set (inode %lu), %s\n", hm->dlopen_file->f_path.dentry->d_inode->i_ino, hm->dlopen_file->f_path.dentry->d_name.name);
            }
        } 
    }

    int vet_sc = 0;
    int vet_sc_symbol = 0;
    unsigned int modreglist = 0;
#ifdef USE_CACHE
    if (!syscall_stored) {
#else
    if (1) {
#endif
        DPRINT("\tSyscall cache miss, vetting\n");
        vet_sc = vet_syscall(syscall_file, syscall_offset, syscall_number, &modreglist);
#if CL == SYMBOLS
        if (syscall_dynamic == 0 && library_dynamic == 0) {
            if (!syscall_stored_vm->badlr) {
                if (library_file == syscall_file) {
                    vet_sc_symbol = vet_symbol(hm->vma, hm->vma_num, subreturn_address, 0);
                } else {
                    vet_sc_symbol = vet_symbol(hm->vma, hm->vma_num, return_address, subreturn_address);
                }
            } else {
                vet_sc_symbol = vet_symbol(hm->vma, hm->vma_num, return_address, 0);
            }
        }
#endif
        if (vet_sc == -3) {
            syscall_stored_vm->badlr = 1;
            vet_sc = 0;
            WARN_BASIC_INFO();
            WARN("Bad subreturn\n");
        }
        if (vet_sc == 0) {
            syscall_stored = syscall_number;
            syscall_stored_vm->addr = return_address;
            syscall_stored_vm->syscall_number = syscall_number;
            syscall_stored_vm->modreglist = modreglist;
        }
    } 
    else {
        if (syscall_stored != syscall_number) {
            vet_sc = -1;
        }
        modreglist = syscall_stored_vm->modreglist;
    }

    if (vet_sc == 0 && vet_sc_symbol == 0) {
        DPRINT("\tSyscall verified\n");
    } else {
        WARN_BASIC_INFO();
        WARN("\tSyscall not verified, error code %d (origin/argument check) and %d (symbol check) -- MC Killa in da house!\n", vet_sc, vet_sc_symbol);
        force_sig(SIGKILL, current);
        return -1;
    }
    
    if (syscall_stored_vm->badlr) {
        WARN("\tCan't vet subreturn %#x LR is used as scratch\n", subreturn_address);
        return 0;
    }
    

    int vet_sr = 0;
    int vet_sr_symbol = 0;
    char err_msg[128];

    if (lr_kill) {
        WARN_BASIC_INFO();
        WARN("\tLR has bad value\n")
        force_sig(SIGKILL, current);
        return -1;
    }

    p = hashget(hm, subreturn_address);
    struct vetmap_t *vm_subreturn_vm = 0;
tryagain2:
    vm_subreturn_vm = hm->hm;
    vm_subreturn_vm = &(vm_subreturn_vm[p]);
#ifdef COLLISION_FREE
    if (vm_subreturn_vm->addr == subreturn_address) {
    }
    else if (!vm_subreturn_vm->addr) {
        vm_subreturn_vm->ts = jiffies;
    }
    else {
        p = (p + 1) % HASHMAP_SIZE;
        goto tryagain2;
    }
#endif

    if (!library_file) {
        vet_sr = 0;
        goto endsubvet;
    }

    if (library_file == syscall_file) {
        argc = 0;
    }

#ifdef USE_CACHE
    if (vm_subreturn_vm->addr != subreturn_address) {
#else
    if (1) {
#endif
        DPRINT("\tSubreturn cache miss, vetting\n");
        vet_sr = vet_subreturn(library_file, library_offset, argv, argc, modreglist, &vm_subreturn_vm);
#if CL == SYMBOLS
        if (vet_sr == 0 && library_dynamic == 0) {
            vet_sr_symbol = vet_symbol(hm->vma, hm->vma_num, subreturn_address, 0);
        }
#endif
        if (vet_sr == 0) {
            vm_subreturn_vm->addr = subreturn_address;
            vm_subreturn_vm->subreturn_valid = vet_sr_symbol == 0 ? 1 : 0;
            vm_subreturn_vm->syscall_number = syscall_number;
            vm_subreturn_vm->syscall_return = return_address;
        }
        if (vet_sr == -2) {
            vm_subreturn_vm->addr = subreturn_address;
            vm_subreturn_vm->subreturn_valid = 0;
            vm_subreturn_vm->syscall_number = 0;
            vm_subreturn_vm->syscall_return = 0;
            vet_sr = 0;
        }
    } else if (vm_subreturn_vm->syscall_number && library_file != syscall_file) {
        if (vm_subreturn_vm->syscall_number != syscall_number) {
            snprintf(err_msg, "\t\tSyscall number from cache does not match: %#x\n", vm_subreturn_vm->syscall_number, sizeof(err_msg));
            vet_sr = -2;
        }

        for (unsigned i = 0; i < argc; i++) {
            if (vm_subreturn_vm->argv[i] != 0xdeadbeef && vm_subreturn_vm->argv[i] != argv[i]) {
                snprintf(err_msg, "\t\tArgument mismatch in A%u -> %x\n", i, vm_subreturn_vm->argv[i], sizeof(err_msg));
                vet_sr = -1;
                break;
            }
        }
#if CL == SYMBOLS
        if (vm_subreturn_vm->syscall_return != return_address) {
            snprintf(err_msg, "\t\tVerified syscall return symbol mismatch\n", sizeof(err_msg));
            vet_sr_symbol = -1;
        }
#endif
    }
endsubvet:
    if (vet_sr == 0 && vet_sr_symbol == 0) {
        DPRINT("\tSubreturn verified\n");
    } else {
        WARN_BASIC_INFO();
        WARN("%s", err_msg);
        WARN("\tSubreturn not verified, error code %d (origin/argument check) and %d (symbol check) -- You will be terminated by M&C\n", vet_sr, vet_sr_symbol);
        force_sig(SIGKILL, current);
        return -1;
    }

    return 0;
}
