/*
 *  BINtegrity aka Syswall
 *
 */

#ifndef __MIPS_DASM_H__
#define __MIPS_DASM_H__

typedef enum {UNKNOWN, ARTH_LOG, LD_ST, JUMP} instruction_type;

typedef enum {ZERO, AT, V0, V1, A0, A1, A2, A3, T0, T1, T2, T3, T4, T5, T6, T7, S0, S1, S2, S3, S4, S5, S6, S7, T8, T9, K0, K1, GP, SP, FP, RA} register_type;

typedef enum {UNDEF, NOP, BLTZ, BGEZ, TGEI, TGEIU, TLTI, TLTIU, TEQI, TNEI, BLTZAL, BGEZAL, MADD, MADDU, MUL, MSUB, MSUBU, CLZ, CLO, SLL, SRL, SRA, SLLV, SRLV, SRAV, JR, JALR, SYSCALL, MFHI, MTHI, MFLO, MTLO, MULT, MULTU, DIV, DIVU, ADD, ADDU, SUB, SUBU, AND, OR, XOR, NOR, SLT, SLTU, J, JAL, BEQ, BNE, BLEZ, BGTZ, ADDI, ADDIU, SLTI, SLTIU, ANDI, ORI, XORI, LUI, LLO, LHI, TRAP, LB, LH, LWL, LW, LBU, LHU, LWR, SB, SH, SWL, SW, SWR, LL, SC} operation_type;

typedef struct {
    instruction_type type;
    operation_type op;
    register_type dst;
    register_type src[2];
    int32_t imm;
    
} mips_instruction_t;

typedef struct {
    int32_t regs[32];
    uint32_t pc;
} reg_state_t;


int mips_disassemble(mips_instruction_t *instruction_buffer, uint32_t number);
int mips_execute(mips_instruction_t* instruction, reg_state_t* reg_state);

static const char * const mips_register_names[32] = {
    "$zero", 
    "$at",
    "$v0",
    "$v1",
    "$a0",
    "$a1",
    "$a2",
    "$a3",
    "$t0",
    "$t1",
    "$t2",
    "$t3",
    "$t4",
    "$t5",
    "$t6",
    "$t7",
    "$s0",
    "$s1",  
    "$s2",
    "$s3",
    "$s4",
    "$s5",
    "$s6",
    "$s7",
    "$t8",
    "$t9",
    "$k0",
    "$k1",
    "$gp",
    "$sp",
    "$fp",
    "$ra" 
};

static const char * const mips_operation_names[128] = {"undef", "nop", "bltz", "bgez", "tgei", "tgeiu", "tlti", "tltiu", "teqi", "tnei", "bltzal", "bgezal", "madd", "maddu", "mul", "msub", "msubu", "clz", "clo", "sll", "srl", "sra", "sllv", "srlv", "srav", "jr", "jalr", "syscall", "mfhi", "mthi", "mflo", "mtlo", "mult", "multu", "div", "divu", "add", "addu", "sub", "subu", "and", "or", "xor", "nor", "slt", "sltu", "j", "jal", "beq", "bne", "blez", "bgtz", "addi", "addiu", "slti", "sltiu", "andi", "ori", "xori", "lui", "llo", "lhi", "trap", "lb", "lh", "lwl", "lw", "lbu", "lhu", "lwr", "sb", "sh", "swl", "sw", "swr", "ll", "sc"};

static operation_type const mips_register_rt_instruction_map[4][8] = {
    {BLTZ, BGEZ},
    {TGEI, TGEIU, TLTI, TLTIU, TEQI, UNDEF, TNEI},
    {BLTZAL, BGEZAL}
};

static operation_type const mips_register_c_instruction_map[8][8] = {
    {MADD, MADDU, MUL, UNDEF, MSUB, MSUBU},
    {},
    {},
    {},
    {CLZ, CLO}
};

static operation_type const mips_register_instruction_map[8][8] = {
    {SLL, UNDEF, SRL, SRA, SLLV, UNDEF, SRLV, SRAV},
    {JR, JALR, UNDEF, UNDEF, SYSCALL},
    {MFHI, MTHI, MFLO, MTLO},
    {MULT, MULTU, DIV, DIVU},
    {ADD, ADDU, SUB, SUBU, AND, OR, XOR, NOR},
    {UNDEF, UNDEF, SLT, SLTU}
};

static operation_type const mips_root_instruction_map[8][8] = {
	{UNDEF, UNDEF, J, JAL, BEQ, BNE, BLEZ, BGTZ},
	{ADDI, ADDIU, SLTI, SLTIU, ANDI, ORI, XORI, LUI},
	{},
	{LLO, LHI, TRAP},
	{LB, LH, LWL, LW, LBU, LHU, LWR},
	{SB, SH, SWL, SW, UNDEF, UNDEF, SWR},
    {LL},
    {SC}
};

#define MIPS_TYPE_R 'R'
#define MIPS_TYPE_J 'J'
#define MIPS_TYPE_I 'I'

#define MIPS_REG_C_TYPE_MULTIPLY 0
#define MIPS_REG_C_TYPE_COUNT 4

#define MIPS_REG_TYPE_SHIFT_OR_SHIFTV 0
#define MIPS_REG_TYPE_JUMPR 1
#define MIPS_REG_TYPE_MOVE 2
#define MIPS_REG_TYPE_DIVMULT 3
#define MIPS_REG_TYPE_ARITHLOG_GTE 4

#define MIPS_ROOT_TYPE_JUMP_OR_BRANCH 0
#define MIPS_ROOT_TYPE_ARITHLOGI 1
#define MIPS_ROOT_TYPE_LOADI_OR_TRAP 3
#define MIPS_ROOT_TYPE_LOADSTORE_GTE 4


#endif
